#!/usr/bin/env python
#  
#  For more information, please see: http://software.sci.utah.edu
#  
#  The MIT License
#  
#  Copyright (c) 2011 Scientific Computing and Imaging Institute,
#  University of Utah.
#  
#  
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#  
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#  
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#  

from sys import argv
from sys import exit
import os
import shutil
import sys
import traceback

# Copy named files to dest
def copy_files( srcs, dest ) :
	
	for file_name in srcs :
		# first try to remove the file with the same name in the
		# destination directory
		
		base_file_name = os.path.basename( file_name )
		dest_fname = os.path.join( dest, base_file_name )
		
		try:
			if ( os.path.exists( dest_fname ) ) :
				os.remove( dest_fname )
		except:
			print( 'Unable to remove %s' % dest_fname )
			exit( 1 )
			
		try:
			shutil.move( file_name, dest )
			print("Moving %s to %s" % ( file_name, dest ) )
		except:
			print( 'Could not move %s to %s' % ( file_name, dest ) )
			exit( 1 )

# grab the outputs from stage "stage_id"
def import_outputs_list( stage_id ) :

	inputs_stage_dir = os.path.join( project_path, stage_id )
	state_fname = os.path.join( inputs_stage_dir, "state.txt" )
	
	if ( not( os.path.exists( state_fname ) ) ) :
		print( 'No such file %s.' % state_fname )
		exit( 1 )
	
	try:
		state_file = open( state_fname ).read()
	except:
		print( 'Failed to open %s\n' % state_fname )
		exit( 1 )
	
	try:
		exec( state_file, globals() )
	except:
		print( 'Failed to exec %s in update.py' % state_fname )
		traceback.print_exc()
		exit( 1 )
	
	# we expect that outputs was defined in the state.txt file
	# it should list all the output files from the previous stage.
	try:
		outputs
		#print('Outputs: ', outputs);
	except NameError:
		print( 'outputs not defined in %s\n' % state_fname )
		exit( 1 )
	
	return outputs

def build_stage_outputs_from_inputs( stage_id, inputs ) :
	outputs = []
	for fname in inputs :
		(base, suffix) = os.path.splitext( fname )
		outputs.append(base+stage_id+suffix)
		
	return outputs

# copy the outputs from the specified stage into the
# output directory.
def copy_outputs_to_stage( outputs, to_stageid ) :

	output_path = os.path.join( project_path, to_stageid )
	
	outputs_full = []

	for file_name in outputs :
		if ( os.path.isabs( file_name ) ) :
			full_file_name = file_name 
		else :
			full_file_name = os.path.join( output_path, file_name )
		if ( not( os.path.isfile( full_file_name ) ) ) :
			print( 'File %s does not exist!' % full_file_name )
			exit( 1 )	
		outputs_full.append( full_file_name )

	copy_files( outputs_full, output_path )

# write out the state.txt file to the target directory.
def update_state( state_fname ) :
	
	state_file = open( state_fname, 'w' )
	
	# write stage outputs
		
	state_file.write( "outputs = %s\n" % outputs )
		
	# stages may have some specific variables they need to 
	# update
	update_state_stage_specific( state_file )
	
	state_file.close()

if __name__ == "__main__" :
	if len( argv ) != 4 :
		print( 'Usage: %s project_path working_dir_path stage_id.' % argv[0] )
		exit( 1 )
	if ( not( os.path.isdir( argv[1] ) ) ) :
		print( 'Project path is not a directory: %s.' % argv[1] )
		exit( 1 )
	if ( not( os.path.isdir( argv[2] ) ) ) :
		print( 'Output path is not a directory: %s.' % argv[2] )
		exit( 1 )

	script_name = argv[0]

	# check for the existence of a parameter file in the input directory.
	project_path = argv[1]
	working_dir_path = argv[2]
	current_stage_id = argv[3]
	
	parameter_fname = os.path.join(working_dir_path,"params.txt")
	if ( not( os.path.exists( parameter_fname ) ) ) :
		print( 'No such file %s.' % parameter_fname )
		exit( 1 )

	scripts_path = os.path.dirname( script_name ) 
	
	# source the parameter.txt file to read in the variables
	# note: this is a security issue if the origin of this file is not 
	# trusted.
	try:
		parameter_file = open( parameter_fname ).read()
	except:
		print( 'Could not open %s for reading.' % parameter_fname )  
		exit( 1 )

	# Run the params.txt file so that its variables get read in.
	# If this fails, the following line will return an exception.
	# We choose not to handle it, as the commandline error is more
	# informative this way.
	try:
		exec( parameter_file )
	except:
		print( 'Failed to exec %s in update.py' % parameter_fname )
		traceback.print_exc()
		exit( 1 )
	
	current_stage_index = 0
	
	valid_stage = False
		
	# find the index of the current stage.
	try:
		for stage_id in stage_ids :
			if ( stage_id == current_stage_id ) :
				valid_stage = True
				break;
			current_stage_index = current_stage_index + 1
	except NameError:
		print( 'stage_ids undefined\n' )
		exit( 1 )
	
	if ( not( valid_stage ) ) :
		print( 'Invalid stage specified %s' % current_stage_id )
		exit( 1 )
	
	if ( current_stage_index < 1 or current_stage_index > 4 ) :
		print( 'Nothing to run for this stage.  Exiting.\n')
		exit( 0 )

	# read in the stage-specific code
	
	stage_specific_fname = os.path.join( scripts_path, current_stage_id+'.py' )
	
	try:
		stage_specific = open( stage_specific_fname ).read()
	except:
		print( 'Could not open %s for reading.' % stage_specific_fname )  
		exit( 1 )
	
	try:
		exec( stage_specific )
	except:
		print( 'Failed to exec %s in update.py' % stage_specific_fname )
		traceback.print_exc()
		exit( 1 )

	outputs = build_stage_outputs_list( )

	outputs_path = build_output_path()	
	
	state_path = os.path.join( outputs_path, "state.txt" )

	# write the updated state
	update_state( state_path )
	
	stage_path = os.path.join( project_path, current_stage_id )

	# copy back the outputs as well as the state file
	stage_results = []	
	for fname in outputs :
		stage_results.append( os.path.join( outputs_path, fname ) )
	
	stage_results.append( state_path )
	
	# copy the outputs back to the permanent directory.
	copy_files( stage_results, stage_path )

	# cleanup -- remove unnecessary files
	cleanup( )
	
